#include <stdio.h>
/* rand */
#include <stdlib.h>
/* time */
#include <time.h>
/* strcmp */
#include <string.h>

#include "array.h"
#include "digital-fov-other.h"
#include "digital-fov.h"
#include "util.h"

#define RADIUS 10
#define NUM_MAPS 100
#define NUM_LOOP 1000
#define WALL_RARITY 6

static int compare_speed(int radius, int num_maps, int num_loop,
                         int wall_rarity,
                         int ignore_fov_brute, int empty_maps);

static int
compare_speed(int radius, int num_maps, int num_loop,
              int wall_rarity,
              int ignore_fov_brute, int empty_maps)
{
  int n;
  int x;
  int y;
  int loop;
  int map_size_x;
  int map_size_y;
  int center_x;
  int center_y;
  time_t t_start;
  time_t t_end;
  int ***maps = NULL;
  int **map_fov = NULL;

  if (radius < 0)
    return 1;
  if (num_maps <= 0)
    return 1;
  if (num_loop <= 0)
    return 1;
  if (wall_rarity <= 0)
    return 1;

  printf("radius = %d, num_maps = %d, num_loop = %d\n",
         radius, num_maps, num_loop);
  if (empty_maps)
    printf("using empty map%s\n", (num_maps <= 1) ? "" : "s");
  else
    printf("wall_rarity = %d\n", wall_rarity);
  if (ignore_fov_brute)
    printf("ignoring digital_fov_brute()\n");
  
  map_size_x = 2 * radius + 1;
  map_size_y = 2 * radius + 1;
  center_x = radius;
  center_y = radius;

  maps = array3_new(num_maps, map_size_x, map_size_y);
  if (maps == NULL)
  {
    fprintf(stderr, "can't allocate memory for maps\n");
    return 1;
  }
  map_fov = array2_new(2 * radius + 1, 2 * radius + 1);
  if (map_fov == NULL)
  {
    fprintf(stderr, "can't allocate memory for map_fov\n");
    array3_delete(maps, num_maps, map_size_x, map_size_y);
    maps = NULL;
    return 1;
  }
  for (n = 0; n < num_maps; n++)
  {
    for (x = 0; x < map_size_x; x++)
    {
      for (y = 0; y < map_size_y; y++)
      {
        if ((!empty_maps) && (rand() % wall_rarity == 0))
          maps[n][x][y] = 1;
        else
          maps[n][x][y] = 0;
      }
    }
  }

  printf("generated maps\n");

  if (!ignore_fov_brute)
  {
    t_start = time(NULL);
    for (loop = 0; loop < num_loop; loop++)
    {
      for (n = 0; n < num_maps; n++)
      {
        digital_fov_brute(maps[n], map_size_x, map_size_y,
                          map_fov, center_x, center_y, radius);
      }
    }
    t_end = time(NULL);
    printf("digital_fov_brute(): %d\n", (int) (t_end - t_start));
  }

  t_start = time(NULL);
  for (loop = 0; loop < num_loop; loop++)
  {
    for (n = 0; n < num_maps; n++)
    {
      digital_fov_beam(maps[n], map_size_x, map_size_y,
                       map_fov, center_x, center_y, radius);
    }
  }
  t_end = time(NULL);
  printf("digital_fov_beam(): %d\n", (int) (t_end - t_start));

  t_start = time(NULL);
  for (loop = 0; loop < num_loop; loop++)
  {
    for (n = 0; n < num_maps; n++)
    {
      digital_fov(maps[n], map_size_x, map_size_y,
                  map_fov, center_x, center_y, radius);
    }
  }
  t_end = time(NULL);
  printf("digital_fov() (recursive): %d\n", (int) (t_end - t_start));

  array3_delete(maps, num_maps, map_size_x, map_size_y);
  maps = NULL;
  array2_delete(map_fov, 2 * radius + 1, 2 * radius + 1);
  map_fov = NULL;

  return 0;
}

int
main(int argc, char *argv[])
{
  int i;
  int radius;
  int num_maps;
  int num_loop;
  int wall_rarity;
  int ignore_fov_brute;
  int empty_maps;

  radius = RADIUS;
  num_maps = NUM_MAPS;
  num_loop = NUM_LOOP;
  wall_rarity = WALL_RARITY;
  ignore_fov_brute = 0;
  empty_maps = 0;

  for (i = 1; i < argc; i++)
  {
    if (!strcmp(argv[i], "--radius"))
    {
      if (i + 1 >= argc)
      {
        fprintf(stderr, "main: missing argument for %s\n", argv[i]);
        return 1;
      }
      if (str_to_int(&radius, argv[i + 1], 0) != 0)
      {
        fprintf(stderr, "main: str_to_int failed\n");
        return 1;
      }
      if (radius < 0)
      {
        fprintf(stderr, "main: radius (%d) must be non-negative\n", radius);
        return 1;
      }
      i++;
    }
    else if (!strcmp(argv[i], "--num-maps"))
    {
      if (i + 1 >= argc)
      {
        fprintf(stderr, "main: missing argument for %s\n", argv[i]);
        return 1;
      }
      if (str_to_int(&num_maps, argv[i + 1], 0) != 0)
      {
        fprintf(stderr, "main: str_to_int failed\n");
        return 1;
      }
      if (num_maps <= 0)
      {
        fprintf(stderr, "main: num_maps (%d) must be positive\n", num_maps);
        return 1;
      }
      i++;
    }
    else if (!strcmp(argv[i], "--num-loop"))
    {
      if (i + 1 >= argc)
      {
        fprintf(stderr, "main: missing argument for %s\n", argv[i]);
        return 1;
      }
      if (str_to_int(&num_loop, argv[i + 1], 0) != 0)
      {
        fprintf(stderr, "main: str_to_int failed\n");
        return 1;
      }
      if (num_loop <= 0)
      {
        fprintf(stderr, "main: num_loop (%d) must be positive\n", num_loop);
        return 1;
      }
      i++;
    }
    else if (!strcmp(argv[i], "--wall-rarity"))
    {
      if (i + 1 >= argc)
      {
        fprintf(stderr, "main: missing argument for %s\n", argv[i]);
        return 1;
      }
      if (str_to_int(&wall_rarity, argv[i + 1], 0) != 0)
      {
        fprintf(stderr, "main: str_to_int failed\n");
        return 1;
      }
      if (wall_rarity <= 0)
      {
        fprintf(stderr, "main: wall_rarity (%d) must be positive\n",
                wall_rarity);
        return 1;
      }
      i++;
    }
    else if (!strcmp(argv[i], "--ignore-fov-brute"))
    {
      ignore_fov_brute = 1;
    }
    else if (!strcmp(argv[i], "--empty-maps"))
    {
      empty_maps = 1;
    }
    else
    {
      fprintf(stderr, "main: unknown option: %s\n", argv[i]);
      return 1;
    }
  }

  srand((unsigned int) time(NULL));

  return compare_speed(radius, num_maps, num_loop,
                       wall_rarity, ignore_fov_brute, empty_maps);
}
