/* FILE */
#include <stdio.h>
/* errno */
#include <errno.h>
/* strerror */
#include <string.h>
/* isupper */
#include <ctype.h>
/* abs */
#include <stdlib.h>

#include "digital-fov.h"

#include "util-example.h"

int
distance(int ax, int ay, int bx, int by)
{
  int dx;
  int dy;

  dx = bx - ax;
  dy = by - ay;

  if (abs(dx) > abs(dy))
    return abs(dx);

  return abs(dy);
}

/* return 1 (true) or 0 (false) */
int
grid_is_illegal(int x, int y, int map_size_x, int map_size_y)
{
  if ((x < 0) || (x >= map_size_x))
    return 1;
  if ((y < 0) || (y >= map_size_y))
    return 1;

  return 0;
}

int
trim_circle(int **map_fov, int radius)
{
  int x;
  int y;
  int y_start;
  int radius2;

  if (map_fov == NULL)
    return 1;

  /* comparing to radius * radius produces an ugly result
   * note that (radius + 0.5)^2 = radius^2 + radius + 0.25
   */
  radius2 = radius * radius + radius;

  /* if abs(x) + abs(y) <= radius, then x^2 + y^2 <= radius^2 */
  for (y_start = 1; y_start <= radius; y_start++)
  {
    x = radius;
    y = y_start;
    while (y <= radius)
    {
      if (x * x + y * y > radius2)
        map_fov[radius + x][radius + y] = 0;
      x--;
      y++;
    }
  }

  for (y_start = 1; y_start <= radius; y_start++)
  {
    x = -radius;
    y = y_start;
    while (y <= radius)
    {
      if (x * x + y * y > radius2)
        map_fov[radius + x][radius + y] = 0;
      x++;
      y++;
    }
  }

  for (y_start = -1; y_start >= -radius; y_start--)
  {
    x = radius;
    y = y_start;
    while (y >= -radius)
    {
      if (x * x + y * y > radius2)
        map_fov[radius + x][radius + y] = 0;
      x--;
      y--;
    }
  }

  for (y_start = -1; y_start >= -radius; y_start--)
  {
    x = -radius;
    y = y_start;
    while (y >= -radius)
    {
      if (x * x + y * y > radius2)
        map_fov[radius + x][radius + y] = 0;
      x++;
      y--;
    }
  }

  return 0;
}

/* Field Of Vision
 * uses Line Of Sight (digital_los())
 * _very_ slow, only for debugging
 * runs at O(N^3), even slower than digital_fov_brute()
 * return 0 on success, 1 on error
 */
int
digital_fov_los(int **map, int map_size_x, int map_size_y,
                int **map_fov,
                int center_x, int center_y, int radius)
{
  int x;
  int y;

  if (map == NULL)
    return 1;
  if (map_fov == NULL)
    return 1;
  if (radius < 0)
    return 1;

  for (x = center_x - radius; x <= center_x + radius; x++)
  {
    for (y = center_y - radius; y <= center_y + radius; y++)
    {
      map_fov[x - center_x + radius][y - center_y + radius]
        = digital_los(map, map_size_x, map_size_y,
                      center_x, center_y, x, y);
    }
  }

  return 0;
}

int
dummy_fov(int **map, int map_size_x, int map_size_y,
          int **map_fov,
          int center_x, int center_y, int radius)
{
  int x;
  int y;

  if (map == NULL)
    return 1;
  if (map_fov == NULL)
    return 1;
  if (radius < 0)
    return 1;

  for (x = center_x - radius; x <= center_x + radius; x++)
  {
    for (y = center_y - radius; y <= center_y + radius; y++)
    {
      if (grid_is_illegal(x, y, map_size_x, map_size_y))
        map_fov[x - center_x + radius][y - center_y + radius] = 0;
      else
        map_fov[x - center_x + radius][y - center_y + radius] = 1;
    }
  }

  return 0;
}

int
read_map_from_file(const char *file_name,
                   int **map, int map_size_x_max, int map_size_y_max,
                   int *map_size_x, int *map_size_y,
                   int *center_x, int *center_y, int **map_char,
                   int quiet)
{
  int c;
  int x;
  int y;
  int beginning_of_line;
  int comment;
  int viewpoint_found;
  FILE *fp = NULL;

  if (file_name == NULL)
    return 1;
  if (map == NULL)
    return 1;
  if (map_size_x_max <= 0)
    return 1;
  if (map_size_y_max <= 0)
    return 1;
  if (map_size_x == NULL)
    return 1;
  if (map_size_y == NULL)
    return 1;

  for (x = 0; x < map_size_x_max; x++)
  {
    for (y = 0; y < map_size_y_max; y++)
    {
      map[x][y] = 0;
      if (map_char != NULL)
        map_char[x][y] = '?';
    }
  }

  *map_size_x = 0;
  *map_size_y = 0;
  if (center_x != NULL)
    *center_x = 0;
  if (center_y != NULL)
    *center_y = 0;

  errno = 0;
  fp = fopen(file_name, "r");
  if (fp == NULL)
  {
    if (!quiet)
    {  
      fprintf(stderr, "read_map_from_file: can't open %s", file_name);
      if (errno != 0)
        fprintf(stderr, " (%s)", strerror(errno));
      fprintf(stderr, "\n");
    }
    return 1;
  }

  x = 0;
  y = 0;
  beginning_of_line = 1;
  comment = 0;
  viewpoint_found = 0;
  while (1 == 1)
  {
    if (x < 0)
    {
      if (!quiet)
      { 
        fprintf(stderr, "read_map_from_file: x overflowed\n");
      }
      fclose(fp);
      fp = NULL;
      return 1;
    }
    if (y < 0)
    {
      if (!quiet)
      { 
        fprintf(stderr, "read_map_from_file: y overflowed\n");
      }
      fclose(fp);
      fp = NULL;
      return 1;
    }

    if (y >= map_size_y_max)
      break;

    c = getc(fp);
    if (c == EOF)
      break;

    if (c == '\n')
    {
      if (!comment)
        y++;
      x = 0;
      beginning_of_line = 1;
      comment = 0;
      continue;
    }

    if (comment)
      continue;

    if ((c == ';') && (beginning_of_line))
    {
      comment = 1;
      beginning_of_line = 0;
      continue;
    }

    if ((x >= 0) && (x < map_size_x_max)
        && (y >= 0) && (y < map_size_y_max))
    {
      if (map_char != NULL)
        map_char[x][y] = c;
      if (c == '@')
      {
        if (center_x != NULL)
          *center_x = x;
        if (center_y != NULL)
          *center_y = y;
        map[x][y] = 0;
        viewpoint_found = 1;
      }
      else if ((c == '#') || (isupper(c)))
      {
        map[x][y] = 1;
      }
      else
      {
        map[x][y] = 0;
      }
      x++;
      if (x > *map_size_x)
        *map_size_x = x;
      if (beginning_of_line)
        (*map_size_y)++;
    }

    beginning_of_line = 0;
  }

  fclose(fp);
  fp = NULL;

  if (!viewpoint_found)
  {
    if (center_x != NULL)
      *center_x = (*map_size_x) / 2;
    if (center_y != NULL)
      *center_y = (*map_size_y) / 2;
  }

  return 0;
}

int
random_map(int **map, int map_size_x_max, int map_size_y_max,
           int *map_size_x, int *map_size_y,
           int *center_x, int *center_y, int **map_char,
           int radius, int wall_rarity)
{
  int x;
  int y;

  if (map == NULL)
    return 1;
  if (map_size_x_max <= 0)
    return 1;
  if (map_size_y_max <= 0)
    return 1;
  if (map_size_x == NULL)
    return 1;
  if (map_size_y == NULL)
    return 1;
  if (radius < 0)
    return 1;
  if (wall_rarity <= 0)
    return 1;
  
  *map_size_x = 2 * radius + 1;
  *map_size_y = 2 * radius + 1;
  if (*map_size_x > map_size_x_max - 1)
    *map_size_x = map_size_x_max - 1;
  if (*map_size_y > map_size_y_max - 1)
    *map_size_y = map_size_y_max - 1;
  if (center_x != NULL)
    *center_x = (*map_size_x) / 2;
  if (center_y != NULL)
    *center_y = (*map_size_y) / 2;

  for (x = 0; x < *map_size_x; x++)
  {
    for (y = 0; y < *map_size_y; y++)
    {
      if (rand() % wall_rarity == 0)
      {
        map[x][y] = 1;
        if (map_char != NULL)
          map_char[x][y] = '#';
      }
      else
      {
        map[x][y] = 0;
        if (map_char != NULL)
          map_char[x][y] = '.';
      }
    }
  }

  return 0;
}
